import re

from bibtexparser.latexenc import latex_to_unicode


def _subr(regex, sub, text):
    """
    Recursive replacement.
    """

    # make substitutions, get the number of substitutions "n"
    text, n = re.subn(regex, sub, text)

    # continue substituting
    if n:
        return _subr(regex, sub, text)

    return text


def remove_wrapping_braces(string: str):
    """
    Remove wrapping "{...}".
    :param string: A string.
    :return: The reformatted string.
    """
    return _subr(r"(\{)(.*)(\})", r"\2", string)


def abbreviate_firstname(name: str, sep: str = " "):
    """
    Reformat a name such that first names are abbreviated to initials, for example::

        de Geus, Thomas Willem Jan ->
        de Geus, T. W. J.

    :param name: The name formatted as "Lastname, firstname otherfirstname"
    :param sep: Separator to place between initials.
    :return: Formatted name.
    """

    if len(name.split(",")) == 1:
        return name

    if len(name.split(",")) > 2:
        raise OSError(f'Unable to interpret name "{name}"')

    match = [
        (re.compile(r"(.*)(\(.*\))", re.UNICODE), r"\1"),
        (
            re.compile(r"([\w][\}]*)([\w0-9\{\}\`\'\"\\\.\^\{]*)", re.UNICODE),
            r"\1.",
        ),
        (re.compile(r"([\w\.][\-]?)([\ ]*)", re.UNICODE), r"\1"),
    ]

    last, first = name.split(",")
    first = latex_to_unicode(first)
    first = first.replace(".", ". ").replace("-", "- ").replace(r"\. ", r"\.") + " "
    names = [latex_to_unicode(i[0]) for i in re.findall(r"([^\s]*)(\s+)", first)][1:]

    for i in range(len(names)):
        for regex, sub in match:
            names[i] = re.sub(regex, sub, names[i])

    return last + ", " + sep.join([rm_unicode(i) for i in names]).upper()


def name2key(name: str):
    """
    Extract last name:
    - Without accents.
    - Without spaces.
    - Starting with a capital letter.
    """

    assert len(name.split(",")) > 1

    last = name.split(",")[0]
    last = rm_accents(last)
    last = last.replace("{", "")
    last = last.replace("}", "")
    last = last[0].upper() + last[1:]

    return last


def protect_math(text: str):
    """
    Protect math mode.

    :param name: The name formatted as "Lastname, firstname otherfirstname"
    """

    # skip text without any math
    if len(text.split(r"{\$}")) < 3:
        return text

    match = [
        (re.compile(r"(\{\\\$\})(.*)(\{\\\$\})", re.UNICODE), r"$\2$"),
        (re.compile(r"(\$\$)(.*)(\$\$)", re.UNICODE), r"$\2$"),
        (re.compile(r"(\$)(.*)(\{\\{\})(.*)(\$)", re.UNICODE), r"\1\2{\4\5"),
        (re.compile(r"(\$)(.*)(\{\\}\})(.*)(\$)", re.UNICODE), r"\1\2}\4\5"),
        (re.compile(r"(\$)(.*)(\{\\_\})(.*)(\$)", re.UNICODE), r"\1\2_\4\5"),
        (re.compile(r"(\$)(.*)(\{\\^\})(.*)(\$)", re.UNICODE), r"\1\2^\4\5"),
        (re.compile(r"(\$)(.*)(\\backslash)(.*)(\$)", re.UNICODE), r"\1\2\\\4\5"),
    ]

    for regex, sub in match:
        text = _subr(regex, sub, text)

    return text


def rm_unicode(text: str):
    """
    Remove unicode.
    """

    # NB list not exhaustive, please extend!
    match = [
        ("ç", r"\c{c}"),
        ("è", r"\`{e}"),
        ("é", r"\'{e}"),
        ("É", r"\'{E}"),
        ("ë", r"\"{e}"),
        ("ô", r"\^{o}"),
        ("ö", r"\"{o}"),
        ("ü", r"\"{y}"),
        ("g̃", r"\~{g}"),
        ("ñ", r"\~{n}"),
        ("İ", r"\.{I}"),
        ("à", r"\'{a}"),
        ("ă", r"\v{a}"),
        ("ř", r"\v{r}"),
        ("–", "--"),
        ("—", "--"),
        ("“", "``"),
        ("”", "''"),
        ("×", r"$\times$"),
    ]

    for ex, sub in match:
        text = text.replace(ex, sub)

    return text


def rm_accents(text: str):
    """
    Remove accents.
    """

    text = rm_unicode(text)

    match = [
        (r"\c{c}", "c"),
        (r"\`{e}", "e"),
        (r"\'{e}", "e"),
        (r"\'{E}", "E"),
        (r"\"{e}", "e"),
        (r"\^{o}", "o"),
        (r"\"{o}", "o"),
        (r"\"{y}", "y"),
        (r"\~{g}", "g"),
        (r"\~{n}", "n"),
        (r"\.{I}", "I"),
        (r"\'{a}", "a"),
        (r"\v{a}", "a"),
        (r"\v{r}", "r"),
        (r"{\"{a}}", "a"),
        (r"{\"{u}}", "u"),
        (r"{\'{c}}", "c"),
        (r"{\'{s}}", "s"),
        (r"{\^{i}}", "i"),
        (r"{\v{s}}", "s"),
        (r"{\v{z}}", "z"),
        (r"\aa", "a"),
        (r"{\o}", "o"),
        (r"'", ""),
        (r" ", ""),
    ]

    for ex, sub in match:
        text = text.replace(ex, sub)

    return text


if __name__ == "__main__":
    pass



import unittest

import numpy as np

import GooseBib as bib


class Test_journals(unittest.TestCase):
    """
    GooseBib.journals
    """

    def test_Journal(self):

        name = "name"
        abbreviation = "abbr"
        acronym = "acro"
        entry = bib.journals.Journal(
            name=name, abbreviation=abbreviation, variations=[name, abbreviation]
        )
        entry = entry.unique()

        assume = dict(
            name=name,
            abbreviation=abbreviation,
        )

        self.assertEqual(assume, dict(entry))

        # overwrite abbreviation

        entry.set_abbreviation(name)
        entry = entry.unique()

        assume = dict(
            name=name,
            variations=[abbreviation],
        )

        self.assertEqual(assume, dict(entry))

        entry.set_abbreviation(abbreviation)
        entry = entry.unique()

        assume = dict(
            name=name,
            abbreviation=abbreviation,
        )

        self.assertEqual(assume, dict(entry))

        # add acronym

        entry.set_acronym(acronym)
        entry = entry.unique()

        assume = dict(
            name=name,
            abbreviation=abbreviation,
            acronym=acronym,
        )

        self.assertEqual(assume, dict(entry))

        # overwrite name

        entry.set_name(abbreviation)
        entry = entry.unique()

        assume = dict(
            name=abbreviation,
            acronym=acronym,
            variations=[name],
        )

        self.assertEqual(assume, dict(entry))

        entry.set_name(name)
        entry.set_abbreviation(abbreviation)
        entry = entry.unique()

        # overwrite variations

        entry.add_variation(name)
        entry.add_variations([abbreviation, acronym])
        entry = entry.unique()

        assume = dict(
            name=name,
            abbreviation=abbreviation,
            acronym=acronym,
        )

        self.assertEqual(assume, dict(entry))

    def test_JournalList(self):

        lst = bib.journals.JournalList()
        lst.names = np.array(["a", "b", "c", "a", "d"])
        lst.index = np.array([0, 0, 3, 2, 2])
        lst.name = -1 * np.array([1, 0, 0, 1, 0])
        lst.abbr = -1 * np.array([0, 1, 0, 0, 0])
        lst.acro = -1 * np.array([0, 0, 0, 0, 1])
        lst._renum()

        m = []
        for i in lst:
            m += [dict(i)]

        n = []
        for i in lst:
            n += [dict(i)]

        mylist = [dict(i) for i in lst.tolist()]

        self.assertEqual(m, n)
        self.assertEqual(m, mylist)

        lst.unique()

        assume = [dict(name="a", abbreviation="b", variations=["d"]), dict(name="c")]

        self.assertEqual(assume, [dict(i) for i in lst.tolist()])

    def test_JournalList_no_force_first(self):

        lst = bib.journals.JournalList()
        lst.names = np.array(["a", "b", "c", "a", "d"])
        lst.index = np.array([0, 0, 3, 2, 2])
        lst.name = -1 * np.array([1, 0, 0, 1, 0])
        lst.abbr = -1 * np.array([0, 1, 0, 0, 0])
        lst.acro = -1 * np.array([0, 0, 0, 0, 1])
        lst._renum()

        m = []
        for i in lst:
            m += [dict(i)]

        n = []
        for i in lst:
            n += [dict(i)]

        mylist = [dict(i) for i in lst.tolist()]

        self.assertEqual(m, n)
        self.assertEqual(m, mylist)

        lst.unique(force_first=False)

        assume = [dict(name="a", abbreviation="b", acronym="d"), dict(name="c")]

        self.assertEqual(assume, [dict(i) for i in lst.tolist()])

    def test_JournalList_map(self):

        lst = bib.journals.JournalList()
        lst.names = np.array(["a", "b", "c", "a", "d"])
        lst.index = np.array([0, 0, 3, 2, 2])
        lst.name = -1 * np.array([1, 0, 0, 1, 0])
        lst.abbr = -1 * np.array([0, 1, 0, 0, 0])
        lst.acro = -1 * np.array([0, 0, 0, 0, 1])
        lst._renum()
        lst.unique()

        self.assertEqual(lst.map2name(["b", "foo", "d"]), ["a", "foo", "a"])
        self.assertEqual(lst.map2abbreviation(["b", "foo", "d"]), ["b", "foo", "b"])
        self.assertEqual(lst.map2acronym(["b", "foo", "d"]), ["a", "foo", "a"])

    def test_JournalList_map_no_force_first(self):

        lst = bib.journals.JournalList()
        lst.names = np.array(["a", "b", "c", "a", "d"])
        lst.index = np.array([0, 0, 3, 2, 2])
        lst.name = -1 * np.array([1, 0, 0, 1, 0])
        lst.abbr = -1 * np.array([0, 1, 0, 0, 0])
        lst.acro = -1 * np.array([0, 0, 0, 0, 1])
        lst._renum()
        lst.unique(force_first=False)

        self.assertEqual(lst.map2name(["b", "foo", "d"]), ["a", "foo", "a"])
        self.assertEqual(lst.map2abbreviation(["b", "foo", "d"]), ["b", "foo", "b"])
        self.assertEqual(lst.map2acronym(["b", "foo", "d"]), ["d", "foo", "d"])

    def test_load(self):

        variations = [
            "Proc. Nat. Acad. Sci.",
            "Proc. Nat. Acad. Sci. U. S. A",
            "Proc. Nat. Acad. Sci. U.S.A",
            "Proc. Nat. Acad. Sci. U.S.A.",
            "Proc. Nat. Acad. Sci. USA",
            "Proc. Natl. Acad. Sci. U. S. A",
            "Proc. Natl. Acad. Sci. U.S.A",
            "Proc. Natl. Acad. Sci. U.S.A.",
            "Proc. Natl. Acad. Sci. USA",
            "Proceedings of the National Academy of Sciences of the United States of America",
            "Proceedings of the National academy of Sciences of the United States of America",
        ]

        expect = [
            {
                "name": "Proceedings of the National Academy of Sciences",
                "abbreviation": "Proc. Natl. Acad. Sci.",
                "acronym": "PNAS",
                "variations": variations,
            }
        ]
        read = bib.journals.load("pnas")
        read = [dict(r) for r in read.tolist()]

        self.assertEqual(expect, read)


if __name__ == "__main__":

    unittest.main()



"""GbibCheckAuthors
  List authors that are possibly stored in more than one way.
  Specifically, the algorithm finds those authors that are written in exactly
  the way after parsing such that first names are abbreviated.

Usage:
  GbibCheckAuthors [options] <input>

Arguments:
  input                   Input BibTeX-file.

Options:
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""
# ==================================================================================================
import os
import re
import sys

import bibtexparser
import docopt

from .. import version

# ======================================== REFORMAT AUTHORS ========================================


def reformatAuthor(text, sep=" "):

    # skip authors that cannot be split in first and last name
    if len(text.split(",")) <= 1:
        return text

    # define find/replace criteria
    match = [
        (re.compile(r"(.*)(\(.*\))", re.UNICODE), r"\1"),
        (re.compile(r"([A-Za-z][\}]?)([\w0-9\{\}\'\"\\\.\^\{]*)", re.UNICODE), r"\1."),
        (re.compile(r"([A-Za-z\.][\-]?)([\ ]*)", re.UNICODE), r"\1"),
        (re.compile(r"([A-Za-z\.][\-]?)([A-Za-z])", re.UNICODE), r"\1" + sep + r"\2"),
    ]

    # extract first and last name
    last, first = text.split(",")

    # update extend all "." with a space, to distinguish initials
    first = first.replace(".", ". ")

    # loop over over find/replace criteria
    for regex, sub in match:
        first = re.sub(regex, sub, first)

    # remove spaces from the beginning and end
    first = first.strip()

    # rejoin first and last name
    return last + ", " + first.upper()


# ================================ CONVERT UNICODE SYMBOLS TO LATEX ================================


def replaceUnicode(text):

    # NB list not exhaustive, please extend!
    match = [
        ("ç", r"\c{c}"),
        ("è", r"\`{e}"),
        ("é", r"\'{e}"),
        ("ë", r"\"{e}"),
        ("ô", r"\^{o}"),
        ("ö", r"\"{o}"),
        ("ü", r"\"{y}"),
        ("g̃", r"\~{g}"),
        ("ñ", r"\~{n}"),
        ("İ", r"\.{I}"),
        ("à", r"\'{a}"),
        ("ă", r"\v{a}"),
        ("ř", r"\v{r}"),
        ("–", "--"),
        ("—", "--"),
        ("“", "``"),
        ("”", "''"),
        ("×", r"$\times$"),
    ]

    # loop over over find/replace criteria
    for ex, sub in match:
        text = text.replace(ex, sub)

    return text


# ========================================== MAIN PROGRAM ==========================================


def main():

    # --------------------------------- parse command line arguments -------------------------------

    # parse command-line options/arguments
    args = docopt.docopt(__doc__, version=version)

    # change keys to simplify implementation:
    # - remove leading "-" and "--" from options
    args = {re.sub(r"([\-]{1,2})(.*)", r"\2", key): args[key] for key in args}
    # - change "-" to "_" to facilitate direct use in print format
    args = {key.replace("-", "_"): args[key] for key in args}
    # - remove "<...>"
    args = {re.sub(r"(<)(.*)(>)", r"\2", key): args[key] for key in args}

    # --------------------------------------- check arguments --------------------------------------

    # check that the BibTeX file exists
    if not os.path.isfile(args["input"]):
        print('"{input:s}" does not exist'.format(**args))
        sys.exit(1)

    # ---------------------------------------- parse bib-file --------------------------------------

    # read
    # ----

    with open(args["input"]) as file:
        bib = bibtexparser.load(file, parser=bibtexparser.bparser.BibTexParser())

    # check entries
    # -------------

    names = {}

    for entry in bib.entries:

        # set author separation
        sep = ""

        # fix author abbreviations
        for key in ["author", "editor"]:
            if key in entry:
                for name in entry[key].split(" and "):
                    fmtName = replaceUnicode(reformatAuthor(name, sep))
                    if fmtName not in names:
                        names[fmtName] = [name]
                    if name not in names[fmtName]:
                        names[fmtName] += [name]

    # filter unique
    names = {name: names[name] for name in names if len(names[name]) > 1}

    # output format
    n = max(len(name) for name in names)

    # print
    for name in sorted(names):
        print(("{0:%ds} : {1:s}" % (n)).format(name, "; ".join(names[name])))



import csv
import io
import os
import re
import shutil
import tempfile
from typing import Union

import click
import git
import numpy as np
import yaml
from numpy.typing import ArrayLike


def get_configdir() -> str:
    """
    Return the config directory.
    """
    return os.path.expanduser("~/.config/GooseBib")


class Journal:
    """
    Simple class to store journal info.

    :param name: Journal's name.
    :param abbreviation: Abbreviation of the journal's name (optional).
    :param acronym: Acronym of the journal's name (optional).
    :param variations: Known variation used for the journal's name, abbreviation, etc. (optional).

    :param index:
        For internal use only. Construction can be simplified by specifying name, abbreviation,
        acronym, and variations as a single list for the ``variations`` option.
        ``index`` than indicates the indices in this list corresponding to
        ``[name, abbreviation, acronym]`` (the same index may be use multiple times if there is no
        abbreviation or acronym).

    :param abbreviation_is_acronym: Use abbreviation as acronym if no acronym is specified.
    """

    def __init__(
        self,
        name: str = None,
        abbreviation: str = None,
        acronym: str = None,
        variations: list[str] = None,
        index: list[int] = None,
        abbreviation_is_acronym: bool = False,
    ):

        if index:
            assert name is None
            assert abbreviation is None
            assert variations is not None
            assert len(index) > 0
            assert max(index) < len(variations)
            self.data = [str(i) for i in variations]
            self.name = index[0]
            self.abbr = index[1]
            self.acro = index[2]
            return

        self.name = 0
        self.abbr = 0
        self.acro = 0
        self.data = []

        if name:
            self.set_name(name)

        if abbreviation:
            self.set_abbreviation(abbreviation, abbreviation_is_acronym and acronym is None)

        if acronym:
            self.set_acronym(acronym)

        if variations:
            self.add_variations(variations)

    def set_name(self, arg):
        """
        (Over)write the journal's name.
        :param arg: Name.
        """
        n = len(self.data)
        if self.abbr == self.name:
            self.abbr = n
        if self.acro == self.name:
            self.acro = n
        self.name = n
        self.data.append(arg)

    def set_abbreviation(self, arg: str, also_acronym: bool = False):
        """
        (Over)write the abbreviation of the journal's name.
        :param arg: Name.
        :param also_acronym: Use also as acronym.
        """
        n = len(self.data)
        if also_acronym:
            self.acro = n
        self.abbr = n
        self.data.append(arg)

    def set_acronym(self, arg: str):
        """
        (Over)write the acronym of the journal's name.
        :param arg: Name.
        """
        self.acro = len(self.data)
        self.data.append(arg)

    def add_variation(self, arg: str):
        """
        Add a variation (does not change the name, abbreviation, or acronym).
        :param arg: Name.
        """
        self.data.append(arg)

    def add_variations(self, arg: list[str]):
        """
        Add a list of variations (does not change the name, abbreviation, or acronym).
        :param arg: Names.
        """
        self.data += arg

    def unique(self):
        """
        In place operation.
        Removes duplicates from list of stored name, abbreviation, acronym, variations.
        Does not change the output in any way.
        """
        data, index = np.unique(self.data, return_inverse=True)
        self.data = [str(d) for d in data]
        self.name = index[self.name]
        self.abbr = index[self.abbr]
        self.acro = index[self.acro]
        return self

    def __add__(self, other):
        """
        Merge two :py:class`Journal` entries. The name, abbreviation, and acronym of the first
        occurring item are prioritised, and are taken from the second occurrence only if not
        present in the first.

        Internally duplicate entries are removed.
        """

        e = Journal()
        n = len(self.data)

        if n == 0:
            e.data = other.data
            e.name = other.name
            e.abbr = other.abbr
            e.acro = other.acro
            return e.unique()

        e.data = self.data + other.data
        e.name = self.name
        e.abbr = self.abbr
        e.acro = self.acro

        if self.abbr == self.name and other.abbr != other.name:
            e.abbr = n + other.abbr

        if self.acro == self.name and other.acro != other.name:
            e.acro = n + other.acro

        return e.unique()

    def __iter__(self):
        """
        Allow ``dict(myjournal)``.
        """

        yield "name", self.data[self.name]

        if self.abbr != self.name:
            yield "abbreviation", self.data[self.abbr]

        if self.acro != self.name:
            yield "acronym", self.data[self.acro]

        idx = np.arange(len(self.data))
        idx = idx[np.logical_not(np.in1d(idx, np.unique([self.name, self.abbr, self.acro])))]

        if idx.size > 0:
            yield "variations", [self.data[i] for i in idx]

    def __repr__(self):
        return str(dict(self))


class JournalList:
    """
    Store journal database as list of journals, which allows efficient handling.

    :param data: Database to map.
    """

    def __init__(self, data: Union[dict[Journal], list[Journal]] = None):

        self.count = 0

        if data is None:
            self.names = []
        else:
            if isinstance(data, dict):
                data = [value for (key, value) in sorted(data.items())]
            self.names = []
            for entry in data:
                self.names += entry.data

        n = len(self.names)
        self.names = np.array(self.names)
        self.index = np.empty(n, dtype=int)
        self.name = np.zeros(n, dtype=int)
        self.abbr = np.zeros(n, dtype=int)
        self.acro = np.zeros(n, dtype=int)

        if data is None:
            return

        i = 0
        for ientry, entry in enumerate(data):
            j = i + len(entry.data)
            self.index[i:j] = ientry
            self.name[i + entry.name] = -1
            self.abbr[i + entry.abbr] = -1
            self.acro[i + entry.acro] = -1
            i = j

        self._renum()

    def _renum(self):
        """
        Internal renumbering.
        Private function: external call only needed after outside manipulation of member variables.
        """

        old = np.sort(np.unique(self.index))
        renum = np.empty(np.max(old) + 1, dtype=int)
        renum[old] = np.arange(old.size)
        self.index = renum[self.index]

    def _map(self, journals: list[str], case_sensitive: bool, field: ArrayLike) -> list[str]:
        """
        Map List of names.

        :param journals: List to map.
        :param case_sensitive: Keep case during look-up.
        :param field: Type identifier to use in case of positive match.
        :return: Input list with mapped items replaced where a positive match was found.
        """

        ret = [i for i in journals]

        if case_sensitive:
            variations = self.names
            search = journals
        else:
            variations = [str(i).lower() for i in self.names]
            search = [i.lower() for i in journals]

        _, v_index, s_index = np.intersect1d(variations, search, return_indices=True)

        for j, s in zip(self.index[v_index], s_index):
            sel = self.index == j
            n = np.argmin(field[sel])
            r = str(self.names[sel][n])
            for i in np.argwhere(np.array(ret) == ret[s]).ravel():
                ret[i] = r

        return ret

    def map2name(self, journals: list[str], case_sensitive: bool = False) -> list[str]:
        """
        Map List of names.

        :param journals: List to map.
        :param case_sensitive: Keep case during look-up.
        :return: Input list with official name replaced where a positive match was found.
        """
        return self._map(journals, case_sensitive, self.name)

    def map2abbreviation(self, journals: list[str], case_sensitive: bool = False) -> list[str]:
        """
        Map List of names.

        :param journals: List to map.
        :param case_sensitive: Keep case during look-up.
        :return: Input list with abbreviation replaced where a positive match was found.
        """
        return self._map(journals, case_sensitive, self.abbr)

    def map2acronym(self, journals: list[str], case_sensitive: bool = False) -> list[str]:
        """
        Map List of names.

        :param journals: List to map.
        :param case_sensitive: Keep case during look-up.
        :return: Input list with acronym replaced where a positive match was found.
        """
        return self._map(journals, case_sensitive, self.acro)

    def unique(self, force_first=True) -> bool:
        """
        Merge journal that have a common entry.
        Note that this applies changes in-place.

        :param force_first:
            Add the second, third, ... duplicates only as name variations, do not change
            name, abbreviation, and acronym of the first duplicate.
            If ``False``, the name, abbreviation, and acronym of all duplicates are considered
            if they are not present in the first duplicate.

        :return: ``True`` if the list was unique, i.e. if no changes were applied.
        """

        names, ifoward, ibackward = np.unique(self.names, return_index=True, return_inverse=True)

        if names.size == self.names.size:
            return False

        self.count = 0
        self.names = names
        renum = self.index[ifoward][ibackward]
        old = self.index[self.index != renum]
        new = renum[self.index != renum]

        for o, n in zip(old, new):
            i = min(o, n)
            j = max(o, n)
            flip = self.index == j

            if force_first:
                self.name[flip] = 0
                self.abbr[flip] = 0
                self.acro[flip] = 0

            self.index[flip] = i

        self.index = self.index[ifoward]
        self.name = self.name[ifoward]
        self.abbr = self.abbr[ifoward]
        self.acro = self.acro[ifoward]

        self._renum()

        return True

    def tolist(self) -> list[dict]:
        """
        Return as list of dictionaries. Same as::

            ret = []

            for i in data:
                ret += [dict(i)]
        """

        ret = []

        for i in range(np.max(self.index) + 1):

            i = self.index == i
            e = Journal(
                variations=self.names[i],
                index=[
                    np.argmin(self.name[i]),
                    np.argmin(self.abbr[i]),
                    np.argmin(self.acro[i]),
                ],
            )
            ret += [e]

        return ret

    def __add__(self, other, force_first=True):
        """
        Merge two journal lists.

        :param force_first:
            Add the second list only as name variations, do not change
            name, abbreviation, and acronym of the first list.
            If ``False``, the name, abbreviation, and acronym of the second list are considered
            if they are not present in the first list.
        """

        ret = JournalList()
        ret.names = np.concatenate((self.names, other.names))
        ret.index = np.concatenate((self.index, other.index + np.max(self.index) + 1))
        ret.name = np.concatenate((self.name, other.name))
        ret.abbr = np.concatenate((self.abbr, other.abbr))
        ret.acro = np.concatenate((self.acro, other.acro))
        ret.unique(force_first)
        return ret

    def __iter__(self):
        """
        Reset counter.
        """
        self.count = 0
        return self

    def __next__(self) -> Journal:
        """
        Return next entry as Journal.
        """

        if self.count > np.max(self.index):
            raise StopIteration

        sel = self.index == self.count
        ret = Journal(
            variations=self.names[sel],
            index=[
                np.argmin(self.name[sel]),
                np.argmin(self.abbr[sel]),
                np.argmin(self.acro[sel]),
            ],
        )

        self.count += 1

        return ret

    def __repr__(self):
        return str([dict(i) for i in self.tolist()])


def _database_tolist(data: Union[dict[Journal], list[Journal]]) -> list[dict]:
    """
    Convert database to list of dictionaries.

    :param data: The database.
    :return: List of dictionaries.
    """

    ret = []

    if isinstance(data, dict):
        for key in sorted(data):
            ret.append(dict(data[key]))
    else:
        for entry in data:
            ret.append(dict(entry))

    return ret


def dump(
    filepath: str,
    data: Union[dict[Journal], list[Journal], JournalList],
    force: bool = False,
):
    r"""
    Dump database to YAML file.

    :param filepath: Filename.
    :param data: The database.
    :param force: Do not prompt to overwrite file.
    """

    dirname = os.path.dirname(filepath)

    if not force:
        if os.path.isfile(filepath):
            if not click.confirm(f'Overwrite "{filepath:s}"?'):
                raise OSError("Cancelled")
        elif not os.path.isdir(dirname) and len(dirname) > 0:
            if not click.confirm(f'Create "{os.path.dirname(filepath):s}"?'):
                raise OSError("Cancelled")

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filepath))

    with open(filepath, "w") as file:
        yaml.dump(_database_tolist(data), file, allow_unicode=True, width=200)


def read(filepath: str, abbreviation_is_acronym: bool = True) -> JournalList:
    """
    Load a database.

    :param filepath: Filename.
    :param abbreviation_is_acronym: Use abbreviation for missing acronym (otherwise title is used).
    """

    if not os.path.isfile(filepath):
        raise OSError(f'"{filepath:s} does not exist')

    with open(filepath) as file:
        ret = yaml.load(file.read(), Loader=yaml.FullLoader)

    return JournalList(
        [Journal(abbreviation_is_acronym=abbreviation_is_acronym, **entry) for entry in ret]
    )


def generate_jabref(*domains):
    """
    Generate a database entry from JabRef.

    :param domains:
        Domain(s) to include in the database. Choose from:
        - acs
        - ams
        - annee-philologique
        - dainst
        - entrez
        - general
        - geology_physics
        - geology_physics_variations
        - ieee
        - ieee_strings
        - lifescience
        - mathematics
        - mechanical
        - medicus
        - meteorology
        - sociology
        - webofscience-dots
        - webofscience

    :return: tood
    """

    alllowed = [
        "acs",
        "ams",
        "annee-philologique",
        "dainst",
        "entrez",
        "general",
        "geology_physics",
        "geology_physics_variations",
        "ieee",
        "ieee_strings",
        "lifescience",
        "mathematics",
        "mechanical",
        "medicus",
        "meteorology",
        "sociology",
        "webofscience-dots",
        "webofscience",
    ]

    domains = [domain.lower() for domain in domains]

    filenames = []
    for domain in domains:
        if domain not in alllowed:
            raise OSError(f'Unknown domain "{domain}"')
        filenames.append(f"journal_abbreviations_{domain}.csv")

    tempdir = tempfile.mkdtemp()
    git.Repo.clone_from("https://github.com/JabRef/abbrv.jabref.org.git", tempdir)

    text = []
    for filename in filenames:
        filepath = os.path.join(tempdir, "journals", filename)
        with open(filepath, newline="") as file:
            text += list(filter(None, file.read().replace("’", "'").replace("–", "-").split("\n")))

    shutil.rmtree(tempdir)

    csvfile = io.StringIO("\n".join(set(text)))
    reader = csv.reader(csvfile, delimiter=";")
    ret = dict()

    for name, abbr in reader:
        if name not in ret:
            ret[name] = Journal(name=name, abbreviation=abbr)
        else:
            ret[name] += Journal(name=name, abbreviation=abbr)

    return ret


def _database_merge(
    db: dict[Journal],
    name: str,
    merge: list[str] = [],
    abbreviation: str = None,
    acronym: str = None,
) -> dict[Journal]:
    """
    Merge several database entries

    :oaram db: The database.
    :param name: The official name to keep (also new key).
    :param merge: The keys to merge.
    :param abbreviation: Overwrite journal abbreviation (optional).
    :param acronym: Overwrite journal acronym (optional).
    :return: The database
    """

    if isinstance(merge, str):
        merge = [merge]

    assert name in db or len(merge) > 0

    if name not in db:
        db[name] = db.pop(merge[0])
        merge = merge[1:]

    for m in merge:
        db[name] += db.pop(m)

    db[name].set_name(name)

    if abbreviation:
        db[name].set_abbreviation(abbreviation)

    if acronym:
        db[name].set_acronym(acronym)

    db[name].unique()

    return db


def generate_default(domain: str):
    """
    Generate default database:

        {
            "Official Journal Name": ``Journal``
            ...
        }

    :param domain: Domain (physics, mechanics, PNAS, PNAS-USA).
    :return: Database.
    """

    domain = domain.lower()

    if domain == "physics":

        # basic list from JabRef

        db = generate_jabref("geology_physics", "geology_physics_variations")

        # merge know aliases / set acronyms / add know variations

        _database_merge(
            db=db,
            name="Proceedings of the National Academy of Sciences of the United States of America",
            abbreviation="Proc. Natl. Acad. Sci. U.S.A.",
            acronym="PNAS",
            merge=[
                "Proceedings of the National academy of Sciences of the United States of America"
            ],
        )

        _database_merge(
            db=db,
            name="Proceedings of the National Academy of Sciences",
            abbreviation="Proc. Natl. Acad. Sci.",
            acronym="PNAS",
        )

        _database_merge(
            db=db,
            name="Physical Review A",
            abbreviation="Phys. Rev. A",
            acronym="PRA",
            merge=["Physical Review A: Atomic, Molecular, and Optical Physics"],
        )

        _database_merge(
            db=db,
            name="Physical Review B",
            abbreviation="Phys. Rev. B",
            acronym="PRB",
            merge=["Physical Review B: Condensed Matter"],
        )

        _database_merge(
            db=db,
            name="Physical Review C",
            abbreviation="Phys. Rev. C",
            acronym="PRC",
            merge=["Physical Review C: Nuclear Physics"],
        )

        _database_merge(
            db=db,
            name="Physical Review D",
            abbreviation="Phys. Rev. D",
            acronym="PRD",
            merge=["Physical Review D: Particles and Fields"],
        )

        base = "Physical Review E:"
        _database_merge(
            db=db,
            name="Physical Review E",
            abbreviation="Phys. Rev. E",
            acronym="PRE",
            merge=[
                f"{base} Statistical Physics, Plasmas, Fluids, and Related Interdisciplinary Topics"
            ],
        )

        _database_merge(
            db=db,
            name="Journal of Physics A",
            abbreviation="J. Phys. A",
            merge=[
                "Journal of Physics A: General Physics",
                "Journal of Physics A: Mathematical and General",
                "Journal of Physics A: Mathematical and Theoretical",
                "Journal of Physics A: Mathematical, Nuclear and General",
            ],
        )

        _database_merge(
            db=db,
            name="Journal of Physics B",
            abbreviation="J. Phys. B",
            merge=[
                "Journal of Physics B: Atomic, Molecular and Optical",
                "Journal of Physics B: Atomic, Molecular and Optical Physics",
                "Journal of Physics B: Atomic and Molecular Physics",
            ],
        )

        _database_merge(
            db=db,
            name="Journal of Physics C",
            abbreviation="J. Phys. C",
            merge=["Journal of Physics C: Solid State Physics"],
        )

        _database_merge(
            db=db,
            name="Journal of Physics D",
            abbreviation="J. Phys. D",
            merge=["Journal of Physics D: Applied Physics"],
        )

        _database_merge(
            db=db,
            name="Journal of Physics E",
            abbreviation="J. Phys. E",
            merge=["Journal of Physics E: Scientific Instruments"],
        )

        _database_merge(
            db=db,
            name="Journal of Physics F",
            abbreviation="J. Phys. F",
            merge=["Journal of Physics F: Metal Physics"],
        )

        _database_merge(
            db=db,
            name="Journal of Physics G",
            abbreviation="J. Phys. G",
            merge=["Journal of Physics G: Nuclear and Particle Physics"],
        )

        _database_merge(
            db=db,
            name="Physica A",
            abbreviation="Phys. A",
            merge=["Physica A: Statistical Mechanics and its Applications"],
        )

        _database_merge(
            db=db,
            name="Physica B",
            abbreviation="Phys. B",
            merge=["Physica B: Condensed Matter"],
        )

        _database_merge(
            db=db,
            name="Physica C",
            abbreviation="Phys. C",
            merge=["Physica C: Superconductivity and its Applications"],
        )

        _database_merge(
            db=db,
            name="Physica D",
            abbreviation="Phys. D",
            merge=["Physica D: Nonlinear Phenomena"],
        )

        _database_merge(
            db=db,
            name="Physica E",
            abbreviation="Phys. E",
            merge=[
                "Physica E-low-dimensional Systems & Nanostructures",
                "Physica E: Low-dimensional Systems and Nanostructures",
            ],
        )

        _database_merge(
            db=db,
            name="Zeitschrift für Physik A",
            abbreviation="Z. Phys. A",
            merge=[
                "Zeitschrift für Physik A: Atoms and Nuclei",
                "Zeitschrift für Physik A Hadrons and Nuclei",
            ],
        )

        _database_merge(
            db=db,
            name="Science in China, Series A: Mathematics",
            abbreviation="Sci. China A",
        )

        _database_merge(
            db=db,
            name="Science in China, Series C: Life Sciences",
            abbreviation="Sci. China C",
        )

        _database_merge(
            db=db,
            name="Science in China, Series D: Earth Sciences",
            abbreviation="Sci. China D",
        )

        _database_merge(
            db=db,
            name="Science in China, Series G: Physics Mechanics and Astronomy",
            abbreviation="Sci. China G",
        )

        _database_merge(
            db=db,
            name="Astrophysics and Space Science",
            abbreviation="Astrophys. Space Sci.",
        )

        _database_merge(
            db=db,
            name="Annalen der Physik",
            abbreviation="Ann. Phys.",
        )

        _database_merge(
            db=db,
            name="European Physical Journal B: Condensed Matter and Complex Systems",
            abbreviation="Eur. Phys. J. B",
        )

        _database_merge(
            db=db,
            name="Czechoslovak Journal of Physics",
            abbreviation="Czech. J. Phys.",
        )

        _database_merge(
            db=db,
            name="General Relativity and Gravitation",
            abbreviation="Gen. Relativ. Gravit.",
        )

        _database_merge(
            db=db,
            name="Comptes Rendus Physique",
            abbreviation="C. R. Physique",
        )

        _database_merge(
            db=db,
            name="Journal of Electronic Materials",
            abbreviation="J. Electron. Mater.",
        )

        _database_merge(
            db=db,
            name="Helvetica Physica Acta",
            abbreviation="Helv. Phys. Acta",
        )

        _database_merge(
            db=db,
            name="Beiträge Zur Physik der Atmosphäre",
            abbreviation="Beitr. Zur Phys. Atmos.",
        )

        _database_merge(
            db=db,
            name="IEEE Transactions on Microwave Theory and Techniques",
            abbreviation="IEEE Trans. Microw. Theory Tech.",
        )

        _database_merge(
            db=db,
            name="Journal of Physics: Condensed Matter",
            abbreviation="J. Phys. Condens. Matter",
        )

        _database_merge(
            db=db,
            name="Journal of Quantitative Spectroscopy & Radiative Transfer",
            abbreviation="J. Quant. Spectrosc. Radiat. Transf.",
        )

        _database_merge(
            db=db,
            name="Journal of Statistical Mechanics: Theory and Experiment",
            abbreviation="J. Stat. Mech.",
        )

        _database_merge(
            db=db,
            name="Materials Research Society Symposium Proceedings",
            abbreviation="Mat. Res. Soc. Symp. Proc.",
        )

        base = "Philosophical Transactions of the Royal Society"
        _database_merge(
            db=db,
            name=f"{base} A: Mathematical, Physical and Engineering Sciences",
            abbreviation="Philos. Trans. R. Soc. A",
        )

        base = "Proceedings of the Royal Society of London"
        _database_merge(
            db=db,
            name=f"{base}, Series A: Mathematical and Physical Sciences",
            abbreviation="Proc. R. Soc. London, Ser. A",
        )

        _database_merge(
            db=db,
            name="Proceedings of the Royal Society of London A",
            abbreviation="Proc. R. Soc. Lond. A",
        )

        _database_merge(
            db=db,
            name="Physical Chemistry Chemical Physics",
            abbreviation="Phys. Chem. Chem. Phys.",
        )

        _database_merge(
            db=db,
            name="Quantum Electronics",
            abbreviation="Quantum Electron.",
        )

        _database_merge(
            db=db,
            name="Transactions of the American Geophysical Union",
            abbreviation="Trans. Am. Geophys. Union",
        )

        _database_merge(
            db=db,
            name="Zeitschrift für Kristallographie",
            abbreviation="Z. für Krist.",
        )

        _database_merge(
            db=db,
            name="Proceedings of SPIE",
            abbreviation="Proc. SPIE",
        )

        _database_merge(
            db=db,
            name="Comptes Rendus de l'Académie des Sciences - Series I - Mathematics",
            abbreviation="C. R. Acad. Sci., Ser. I",
        )

        _database_merge(
            db=db,
            name="Comptes Rendus Hebdomadaires des Séances de l'Académie des Sciences",
            abbreviation="C. R. Acad. Sci.",
        )

        _database_merge(
            db=db,
            name="Journal de Physique IV",
            abbreviation="J. Phys. IV",
        )

        _database_merge(
            db=db,
            name="Zeitschrift für Physik",
            abbreviation="Z. Phys.",
        )

        _database_merge(
            db=db,
            name="IEEE Journal of Selected Topics in Quantum Electronics",
            abbreviation="IEEE J. Sel. Top. Quantum Electron.",
        )

        _database_merge(
            db=db,
            name="Phase Transitions",
            abbreviation="Phase Transit.",
        )

        _database_merge(
            db=db,
            name="Nano Futures",
            abbreviation="Nano Futur.",
        )

        _database_merge(
            db=db,
            name="Kongelige Danske Videnskabernes Selskab, Matematisk-Fysiske Meddelelser",
            abbreviation="K. Dan. Vidensk. Selsk. Mat. Fys. Medd.",
        )

        _database_merge(
            db=db,
            name="Modelling and Simulation in Materials Science and Engineering",
            abbreviation="Model. Simul. Mater. Sci. Eng.",
        )

        db["Physical Review Letters"].set_acronym("PRL")
        db["Physical Review X"].set_acronym("PRX")
        db["Journal of the Mechanics and Physics of Solids"].set_acronym("JMPS")
        db["International Journal of Solids and Structures"].set_acronym("IJSS")
        db["Science"].add_variation("Science (80-. ).")
        db["EPL (Europhysics Letters)"].add_variation("EPL (Europhysics Lett.")

    elif domain == "mechanics":

        db = generate_jabref("mechanical")

        _database_merge(
            db=db,
            name="Proceedings of the National Academy of Sciences of the United States of America",
            abbreviation="Proc. Natl. Acad. Sci. U.S.A.",
            acronym="PNAS",
            merge=[
                "Proceedings of the National academy of Sciences of the United States of America"
            ],
        )

        base = "Comptes Rendus de"
        _database_merge(
            db=db,
            name=f"{base} l'Academie des Sciences Serie III: Sciences de la Vie",
            abbreviation="C. R. Acad. Sci., Ser. III",
            merge=[f"{base} l' Academie des Sciences Serie III: Sciences de la Vie"],
        )

        _database_merge(
            db=db,
            name=f"{base} l'Academie des Sciences Serie IIa: Sciences de la Terre et des Planets",
            abbreviation="C. R. Acad. Sci., Ser. IIa",
            merge=[
                f"{base} l' Academie des Sciences Serie IIa:Sciences de la Terre et des Planets"
            ],
        )

        _database_merge(
            db=db,
            name=f"{base} l'Academie des Sciences Serie IIb: Mecanique Physique Chimie Astronomie",
            abbreviation="C. R. Acad. Sci., Ser. IIb",
            merge=[
                f"{base} l' Academie des Sciences Serie IIb:Mecanique Physique Chimie Astronomie"
            ],
        )

        _database_merge(
            db=db,
            name="Comptes Rendus de l'Academie des Sciences Serie IIc: Chemie",
            abbreviation="C. R. Acad. Sci., Ser. IIc",
            merge=["Comptes Rendus de l' Academie des Sciences Serie IIc:Chemie"],
        )

        base = "Proceedings of the Institution of Mechanical Engineers"
        _database_merge(
            db=db,
            name=f"{base}, Part A: Journal of Power and Energy",
            abbreviation="Proc. Inst. Mech. Eng. Part A",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part B: Journal of Engineering Manufacture",
            abbreviation="Proc. Inst. Mech. Eng. Part B",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part C: Journal of Mechanical Engineering Science",
            abbreviation="Proc. Inst. Mech. Eng. Part C",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part D: Journal of Automobile Engineering",
            abbreviation="Proc. Inst. Mech. Eng. Part D",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part E: Journal of Process Mechanical Engineering",
            abbreviation="Proc. Inst. Mech. Eng. Part E",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part F: Journal of Rail and Rapid Transit",
            abbreviation="Proc. Inst. Mech. Eng. Part F",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part G: Journal of Aerospace Engineering",
            abbreviation="Proc. Inst. Mech. Eng. Part G",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part H: Journal of Engineering in Medicine",
            abbreviation="Proc. Inst. Mech. Eng. Part H",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part J: Journal of Engineering Tribology",
            abbreviation="Proc. Inst. Mech. Eng. Part J",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part K: Journal of Multi-body Dynamics",
            abbreviation="Proc. Inst. Mech. Eng. Part K",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part L: Journal of Materials: Design and Applications",
            abbreviation="Proc. Inst. Mech. Eng. Part L",
        )

        _database_merge(
            db=db,
            name=f"{base}, Part M: Journal of Engineering for the Maritime Environment",
            abbreviation="Proc. Inst. Mech. Eng. Part M",
        )

        _database_merge(
            db=db,
            name=f"{base}. Part I: Journal of Systems and Control Engineering",
            abbreviation="Proc. Inst. Mech. Eng. Part I",
        )

        db["International Journal for Numerical Methods in Engineering"].set_acronym("IJNME")
        db["Journal of the Mechanics and Physics of Solids"].set_acronym("JMPS")
        db["International Journal of Solids and Structures"].set_acronym("IJSS")

        key = "Materials Science and Engineering A"
        db[key].add_variation("Materials Science and Engineering: A")
        db[key].set_abbreviation("Mater. Sci. Eng. A")

        db["Procedia Materials Science"] = Journal(
            name="Procedia Materials Science", abbreviation="Procedia Mater. Sci."
        )

    elif domain == "pnas":

        db = generate_default("physics")
        name = "Proceedings of the National Academy of Sciences"
        alias = "Proceedings of the National Academy of Sciences of the United States of America"
        db = {name: db[name] + db[alias]}
        db[name].add_variation("Proc. Nat. Acad. Sci.")
        db[name].add_variation("Proc. Natl. Acad. Sci.")
        db[name].add_variation("Proc. Nat. Acad. Sci. USA")
        db[name].add_variation("Proc. Natl. Acad. Sci. USA")
        db[name].add_variation("Proc. Nat. Acad. Sci. U.S.A")
        db[name].add_variation("Proc. Natl. Acad. Sci. U.S.A")
        db[name].add_variation("Proc. Nat. Acad. Sci. U. S. A")
        db[name].add_variation("Proc. Natl. Acad. Sci. U. S. A")
        db[name].unique()

    elif domain == "pnas-usa":

        db = generate_default("physics")
        name = "Proceedings of the National Academy of Sciences of the United States of America"
        alias = "Proceedings of the National Academy of Sciences"
        db = {name: db[name] + db[alias]}
        db[name].add_variation("Proc. Nat. Acad. Sci.")
        db[name].add_variation("Proc. Natl. Acad. Sci.")
        db[name].add_variation("Proc. Nat. Acad. Sci. USA")
        db[name].add_variation("Proc. Natl. Acad. Sci. USA")
        db[name].add_variation("Proc. Nat. Acad. Sci. U.S.A")
        db[name].add_variation("Proc. Natl. Acad. Sci. U.S.A")
        db[name].add_variation("Proc. Nat. Acad. Sci. U. S. A")
        db[name].add_variation("Proc. Natl. Acad. Sci. U. S. A")
        db[name].unique()

    elif domain == "arxiv":

        db = dict(arXiv=Journal(name="arXiv preprint", abbreviation="arXiv"))

        r = generate_default("physics")
        db["arXiv"] += r["arXiv.org, e-Print Archive Astrophysics"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Condensed Matter"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Physics"]

        r = generate_default("mechanics")
        db["arXiv"] += r["arXiv.org, e-Print Archive Astrophysics"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Condensed Matter"]
        db["arXiv"] += r["arXiv.org, e-Print Archive General Relativity and Quantum Cosmology"]
        db["arXiv"] += r["arXiv.org, e-Print Archive High Energy Physics - Experimental"]
        db["arXiv"] += r["arXiv.org, e-Print Archive High Energy Physics - Lattice"]
        db["arXiv"] += r["arXiv.org, e-Print Archive High Energy Physics - Phenomenology"]
        db["arXiv"] += r["arXiv.org, e-Print Archive High Energy Physics - Theory"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Nuclear Experiment"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Nuclear Theory"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Physics"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Quantitative Biology"]
        db["arXiv"] += r["arXiv.org, e-Print Archive Quantum Physics"]

    else:
        raise OSError(f'Unknown domain "{domain}"')

    return db


def update_default():
    """
    Update the default databases shipped with GooseBib.
    """

    dirname = os.path.dirname(__file__)

    for name in ["physics", "mechanics", "PNAS", "PNAS-USA", "arXiv"]:
        db = JournalList(generate_default(name))
        dump(os.path.join(dirname, f"{name}.yaml"), db, force=True)


def _get_yaml_files(dirname: str) -> list[str]:
    """
    Return YAML-files in a directory.

    :param dirname: Directory to search.
    :return: List of files.
    """

    ret = [f for f in os.listdir(dirname)]
    return [f for f in ret if re.match(r"(\.y)([a]?)(ml)", os.path.splitext(f)[1])]


def load(*args: str) -> JournalList:
    """
    Load database(s) from files (in default locations).
    Note that the order matters: in case of duplicates the first match is leading.

    Custom databases: Store a YAML file to::

        dirname = GooseBib.get_configdir()
        stylename = "mystyle"
        filepath = os.path.join(dirname, f"{stylename}.yaml")

    To file should have the following structure::

        - name: Proceedings of the National Academy of Sciences
          abbreviation: Proc. Natl. Acad. Sci.
          acronym: PNAS
          variations:
          - Proc. Nat. Acad. Sci. U.S.A.
        - name: ...

    whereby all fields except "name" are optional.
    Files stored in ``GooseBib.get_configdir()`` are prioritised over default files shipped
    with the library.

    :param args: physics, mechanics, PNAS, PNAS-USA, ...
    :return: JournalList
    """

    assert len(args) > 0

    search = dict(
        config={"dirname": os.path.abspath(get_configdir())},
        default={"dirname": os.path.abspath(os.path.dirname(__file__))},
    )

    if os.path.exists(search["config"]["dirname"]):
        search_order = ["config", "default"]
    else:
        search_order = ["default"]
        search.pop("config")

    for key in search:
        search[key]["files"] = _get_yaml_files(search[key]["dirname"])
        search[key]["names"] = [os.path.splitext(f)[0].lower() for f in search[key]["files"]]

    db = []

    for arg in args:

        found = False
        name = arg.lower()

        for key in search_order:
            if name in search[key]["names"]:
                names = np.array(search[key]["names"])
                path = search[key]["files"][np.argwhere(names == name).ravel()[0]]
                db += [read(os.path.join(search[key]["dirname"], path))]
                found = True
                break

        if not found:
            raise OSError(f'Could not find "{arg}"')

    ret = db[0]
    for d in db[1:]:
        ret += d
    return ret


if __name__ == "__main__":

    update_default()



"""GbibSelect
  Select only those entries present in a TeX-file.

Usage:
  GbibSelect [options] <tex>...

Options:
  -o, --output=<FILE>   Name of the output BibTeX file.
  -b, --bib=<FILE>      Name of the input BibTeX file.
      --version         Show version.
  -h, --help            Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""
# ==================================================================================================
import os
import re
import sys

import bibtexparser
import docopt

from .. import version

# ==================================== RAISE COMMAND LINE ERROR ====================================


def Error(msg, exit_code=1):

    print(msg)

    sys.exit(exit_code)


# ===================================== EXTRACT CITATION KEYS ======================================


def tex2cite(tex):

    # extract keys from "cite"
    def extract(s):
        try:
            return list(
                re.split(r"([pt])?(\[.*\]\[.*\])?(\{[a-zA-Z0-9\,\-\ ]*\})", s)[3][1:-1].split(",")
            )
        except:
            print("Error in interpreting\n {0} ...").format(s[:100])

    # read all keys in "cite", "citet", "citep" commands
    cite = [extract(i) for i in tex.split(r"\cite")[1:]]
    cite = list({item for sublist in cite for item in sublist})
    cite = [i.replace(" ", "") for i in cite]

    return cite


# ========================================== MAIN PROGRAM ==========================================


def main():

    # --------------------------------- parse command line arguments -------------------------------

    # parse command-line options/arguments
    args = docopt.docopt(__doc__, version=version)

    # change keys to simplify implementation:
    # - remove leading "-" and "--" from options
    args = {re.sub(r"([\-]{1,2})(.*)", r"\2", key): args[key] for key in args}
    # - change "-" to "_" to facilitate direct use in print format
    args = {key.replace("-", "_"): args[key] for key in args}
    # - remove "<...>"
    args = {re.sub(r"(<)(.*)(>)", r"\2", key): args[key] for key in args}

    # --------------------------------------- check arguments --------------------------------------

    # check that the TeX files exist
    for fname in args["tex"]:
        if not os.path.isfile(fname):
            Error(f'"{fname:s}" does not exist')

    # check that the BibTeX file exists
    if args["bib"]:
        if not os.path.isfile(args["bib"]):
            Error('"{:s}" does not exist'.format(args["bib"]))

    # if the file is an existing directory -> convert to file with same name as the input file
    if args["output"]:
        if os.path.isdir(args["output"]):
            args["output"] = os.path.join(args["output"], os.path.split(args["bib"])[-1])

    # -------------------------------------- get citation keys -------------------------------------

    # initialize
    keys = []

    # read from TeX files
    for fname in args["tex"]:
        keys += tex2cite(open(fname).read())

    # sort citation keys
    keys = sorted(list(set(keys)))

    # optionally: print and quit
    if not args["bib"] and not args["output"]:
        print("\n".join(keys))

    # ------------------------------------- select from bib-file -----------------------------------

    # read BibTeX file
    bib = bibtexparser.load(open(args["bib"]), parser=bibtexparser.bparser.BibTexParser())

    # get citation keys in the BibTeX files
    bibkeys = [entry["ID"] for entry in bib.entries]

    # find entries that are present in the TeX files, but missing in the BibTeX files
    missing = [key for key in keys if key not in bibkeys]

    # print missing entries
    if len(missing) > 0:
        Error("Not found: " + ", ".join(missing))

    # select entries based on the keys found in the TeX file
    bib.entries = [entry for entry in bib.entries if entry["ID"] in keys]

    # convert to text
    bib = bibtexparser.dumps(bib)

    # write bib-file, or print to screen
    if args["output"]:
        open(args["output"], "w").write(bib)
    else:
        print(bib)



import unittest

import GooseBib as bib


class Test_tex(unittest.TestCase):
    """
    GooseBib.tex
    """

    def test_list_cise(self):

        key = "DeGeus2021"
        mytext = r"The authors of \cite{DeGeus2021} claim that ..."
        self.assertEqual(bib.tex.list_cite(mytext), [key])


if __name__ == "__main__":

    unittest.main()



"""GbibCheckKeys
  List entries that do not have a citation key composed of "LastName + Year".

Usage:
  GbibCheckKeys [options] <input>

Arguments:
  input                   Input BibTeX-file.

Options:
  -s, --view-skipped      View skipped entries that cannot be composed in the standard way.
      --version           Show version.
  -h, --help              Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""
# ==================================================================================================
import os
import re
import sys

import bibtexparser
import docopt

from .. import version

# ==================================================================================================


def replaceUnicode(text):

    match = [
        ("ç", "c"),
        ("è", "e"),
        ("é", "e"),
        ("ë", "e"),
        ("ô", "o"),
        ("ö", "o"),
        ("ü", "y"),
        ("g̃", "g"),
        ("ñ", "n"),
        ("İ", "I"),
        ("à", "a"),
        ("ă", "a"),
        ("ř", "r"),
        (r"\c{c}", "c"),
        (r"\`{e}", "e"),
        (r"\'{e}", "e"),
        (r"\"{e}", "e"),
        (r"\^{o}", "o"),
        (r"\"{o}", "o"),
        (r"\"{y}", "y"),
        (r"\~{g}", "g"),
        (r"\~{n}", "n"),
        (r"\.{I}", "I"),
        (r"\'{a}", "a"),
        (r"\v{a}", "a"),
        (r"\v{r}", "r"),
        (r"{\"{a}}", "a"),
        (r"{\"{u}}", "u"),
        (r"{\'{c}}", "c"),
        (r"{\'{s}}", "s"),
        (r"{\^{i}}", "i"),
        (r"{\v{s}}", "s"),
        (r"{\v{z}}", "z"),
        (r"\aa", "a"),
        (r"{\o}", "o"),
        (r"'", ""),
        (r" ", ""),
    ]

    for ex, sub in match:
        text = text.replace(ex, sub)

    return text


# ==================================================================================================


def getPlainLastName(full_name):

    last = full_name.split(",")[0]

    last = replaceUnicode(last)
    last = last.replace("{", "")
    last = last.replace("}", "")
    last = last[0].upper() + last[1:]

    return last


# ========================================== MAIN PROGRAM ==========================================


def main():

    # parse command-line options/arguments
    # change keys to simplify implementation:
    # - remove leading "-" and "--" from options
    # - change "-" to "_" to facilitate direct use in print format
    # - remove "<...>"
    args = docopt.docopt(__doc__, version=version)
    args = {re.sub(r"([\-]{1,2})(.*)", r"\2", key): args[key] for key in args}
    args = {key.replace("-", "_"): args[key] for key in args}
    args = {re.sub(r"(<)(.*)(>)", r"\2", key): args[key] for key in args}

    if not os.path.isfile(args["input"]):
        print('"{input:s}" does not exist'.format(**args))
        sys.exit(1)

    alphabet = "abcdefghijklmnopqrstuvwxyz"
    alphabet_list = list(alphabet)

    # get all keys and auto-generated equivalent

    with open(args["input"]) as file:
        bib = bibtexparser.load(file, parser=bibtexparser.bparser.BibTexParser())

    keys = []

    for entry in bib.entries:

        old_key = entry["ID"]
        new_key = entry["ID"]
        generated = False

        if "author" in entry and "year" in entry:
            name = getPlainLastName(entry["author"].split(" and ")[0])
            year = entry["year"]
            new_key = name + year
            generated = True

        if not generated:
            if "editor" in entry and "year" in entry:
                name = getPlainLastName(entry["editor"].split(" and ")[0])
                year = entry["year"]
                new_key = name + year
                generated = True

        keys += [[old_key, new_key, generated]]

    keys = sorted(keys, key=lambda x: x[0])

    # handle duplicate keys by adding (a, b, c, ...); respect old

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):
        if old_key != new_key:
            if ntimes[new_key] > 1:
                if re.match(r"[a-z]", old_key[-1]):
                    new_key += old_key[-1]
                    keys[idx][1] = new_key

    # handle duplicate keys by adding (a, b, c, ...)

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):
        if not generated:
            continue
        if ntimes[new_key] > 1:
            ntimes[new_key] -= 1
            if not re.match(r"[a-z]", new_key[-1]):
                new_key += "a"
            if new_key in ntimes:
                new_key = new_key[:-1] + alphabet_list[alphabet.find(new_key[-1]) + 1]
            keys[idx][1] = new_key
            ntimes[new_key] = 1

    # make sure that duplicate suffix (a, b, c, ...) is as low as possible

    ntimes = {new_key: 0 for _, new_key, _ in keys}

    for old_key, new_key, generated in keys:
        ntimes[new_key] += 1

    for idx, (old_key, new_key, generated) in enumerate(keys):

        if not generated:
            continue

        if not re.match(r"[a-z]", new_key[-1]):
            continue

        if new_key[:-1] not in ntimes:
            keys[idx][1] = new_key[:-1]
            ntimes[new_key[:-1]] = 1
            ntimes[new_key] -= 1
            continue

        if ntimes[new_key[:-1]] == 0:
            keys[idx][1] = new_key[:-1]
            ntimes[new_key[:-1]] += 1
            ntimes[new_key] -= 1
            continue

        for j in range(alphabet.find(new_key[:-1])):
            if new_key[:-1] + alphabet[j] not in ntimes:
                keys[idx][1] = new_key[:-1] + alphabet[j]
                ntimes[keys[idx][1]] = 1
                ntimes[new_key] -= 1
                break

            if ntimes[new_key[:-1] + alphabet[j]] == 0:
                keys[idx][1] = new_key[:-1] + alphabet[j]
                ntimes[keys[idx][1]] += 1
                ntimes[new_key] -= 1
                break

    # print respect

    n = max(len(old_key) for old_key, _, _ in keys)
    fmt = "{0:%ds} : {1:s}" % (n)

    for old_key, new_key, generated in keys:
        if old_key != new_key:
            print(fmt.format(old_key, new_key))

    if args["view_skipped"]:
        skipped = [new_key for old_key, new_key, generated in keys if not generated]
        if len(skipped) > 0:
            print("")
            print("Skipped keys:")
            print("\n".join(skipped))



from setuptools import find_packages
from setuptools import setup

setup(
    name="GooseBib",
    license="MIT",
    author="Tom de Geus",
    author_email="tom@geus.me",
    description="Reformat BibTeX files",
    long_description="Reformat BibTeX files",
    keywords="LaTeX; BibTeX",
    url="https://github.com/tdegeus/GooseBib",
    packages=find_packages(),
    package_data={"": ["*.yaml"]},
    use_scm_version={"write_to": "GooseBib/_version.py"},
    setup_requires=["setuptools_scm"],
    install_requires=[
        "bibtexparser",
        "click",
        "docopt",
        "GitPython",
        "numpy",
        "PyYAML",
        "requests",
    ],
    entry_points={
        "console_scripts": [
            "GbibCheckAuthors = GooseBib.cli.GbibCheckAuthors:main",
            "GbibCheckKeys = GooseBib.cli.GbibCheckKeys:main",
            "GbibCheckLink = GooseBib.cli.GbibCheckLink:main",
            "GbibClean = GooseBib.bibtex:GbibClean",
            "GbibList = GooseBib.cli.GbibList:main",
            "GbibParse = GooseBib.cli.GbibParse:main",
            "GbibSelect = GooseBib.cli.GbibSelect:main",
            "GbibSelectAlias = GooseBib.cli.GbibSelectAlias:main",
            "GbibShowAuthorRename = GooseBib.bibtex:GbibShowAuthorRename",
        ]
    },
)



"""GbibSelectAlias
  Select specific BibTeX fields and (optionally) choose the citation key.
  The selection is provided to the program using a json-file.

  ```json
  {
    "new_key" : "old_key",
    ...
  }
  ```

Usage:
  GbibSelectAlias [options] <json>

Options:
  -o, --output=<FILE>   Name of the output BibTeX file.
  -b, --bib=<FILE>      Name of the input BibTeX file.
      --version         Show version.
  -h, --help            Show help.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
"""
# ==================================================================================================
import json
import os
import re
import sys

import bibtexparser
import docopt

from .. import version

# ==================================== RAISE COMMAND LINE ERROR ====================================


def Error(msg, exit_code=1):

    print(msg)

    sys.exit(exit_code)


# ========================================== MAIN PROGRAM ==========================================

# ---------------------------------- parse command line arguments ----------------------------------

# parse command-line options/arguments
args = docopt.docopt(__doc__, version=version)

# change keys to simplify implementation:
# - remove leading "-" and "--" from options
args = {re.sub(r"([\-]{1,2})(.*)", r"\2", key): args[key] for key in args}
# - change "-" to "_" to facilitate direct use in print format
args = {key.replace("-", "_"): args[key] for key in args}
# - remove "<...>"
args = {re.sub(r"(<)(.*)(>)", r"\2", key): args[key] for key in args}

# ---------------------------------------- check arguments -----------------------------------------

# check that the TeX files exist
if not os.path.isfile(args["json"]):
    Error('"{:s}" does not exist'.format(args["json"]))

# check that the BibTeX file exists
if args["bib"]:
    if not os.path.isfile(args["bib"]):
        Error('"{:s}" does not exist'.format(args["bib"]))

# if the file is an existing directory -> convert to file with same name as the input file
if args["output"]:
    if os.path.isdir(args["output"]):
        args["output"] = os.path.join(args["output"], os.path.split(args["bib"])[-1])

# -------------------------------------- select from bib-file --------------------------------------

# data
with open(args["json"]) as json_file:
    data = json.load(json_file)

# inverse data
data_inv = {data[key]: key for key in data}

# extract lists
keys_new = [key for key in sorted(data)]
keys_old = [data[key] for key in sorted(data)]

# read BibTeX file
bib = bibtexparser.load(open(args["bib"]), parser=bibtexparser.bparser.BibTexParser())

# get citation keys in the BibTeX files
bibkeys = [entry["ID"] for entry in bib.entries]

# find entries that are present in the TeX files, but missing in the BibTeX files
missing = [key for key in keys_old if key not in bibkeys]

# print missing entries
if len(missing) > 0:
    Error("Not found: " + ", ".join(missing))

# select entries based on the keys found in the TeX file
bib.entries = [entry for entry in bib.entries if entry["ID"] in keys_old]

# modify citation keys
for entry in bib.entries:
    entry["ID"] = data_inv[entry["ID"]]

# convert to text
bib = bibtexparser.dumps(bib)

# write bib-file, or print to screen
if args["output"]:
    open(args["output"], "w").write(bib)
else:
    print(bib)



import argparse
import difflib
import inspect
import io
import os
import re
import textwrap
import warnings
from collections import OrderedDict
from functools import singledispatch
from typing import Union

import bibtexparser
import click
import numpy as np

from . import journals
from . import recognise
from . import reformat
from ._version import version


def read_display_order(bibtex_str: str, tabsize: int = 2) -> (dict, int):
    """
    Read order of fields of all entries.

    :param bibtex_str: A BibTeX 'file'.
    :param tabsize: Replace "\t" by a number of spaces.
    :return:
        A dictionary with a list of fields per key.
        The typical indentation.
    """

    ret = {}
    indent = []

    matches = list(re.finditer(r"(\@\w*\{)", bibtex_str, re.I))

    for m in range(len(matches)):

        i = matches[m].start()

        if m < len(matches) - 1:
            j = matches[m + 1].start()
            entry = bibtex_str[i:j]
        else:
            entry = bibtex_str[i:]

        components = re.split(r"(.*)(\@\w*\{)", entry)

        if components[2].lower() in ["@string{", "@comment{", "@preamble{"]:
            continue

        key, data = components[3].split(",", 1)
        find = re.findall(r"([\n\t\ ]*)([\w\_\-]*)([\ ]?=)(.*)", data)
        ret[key] = [i[1] for i in find]
        indent += [len("".join(i[0].replace("\t", tabsize * " ").splitlines())) for i in find]

    if len(indent) == 0:
        indent = 0
    else:
        indent = int(np.ceil(np.mean(indent)))

    return ret, indent


class MyBibTexWriter(bibtexparser.bwriter.BibTexWriter):
    """
    Overload of ``bibtexparser.bwriter.BibTexWriter`` acting on an extra internal field
    "DISPLAY_ORDER" to preserve the order of each item.
    """

    def __init__(self, *args, **kwargs):
        sort = kwargs.pop("sort_entries", False)
        super().__init__(self, *args, **kwargs)
        if not sort:
            self.order_entries_by = []

    def _entry_to_bibtex(self, entry):
        self.display_order = entry.pop("DISPLAY_ORDER", [])
        self.indent = entry.pop("INDENT", " ")
        return bibtexparser.bwriter.BibTexWriter._entry_to_bibtex(self, entry)

    def write(self, *args, **kwargs):
        ret = bibtexparser.bwriter.BibTexWriter.write(self, *args, **kwargs)
        return ret.rstrip() + "\n"


class MyBibTexParser(bibtexparser.bparser.BibTexParser):
    """
    Overload of ``bibtexparser.bparser.BibTexParser`` adding an extra internal field
    "DISPLAY_ORDER" to preserve the order of each item.
    """

    def parse(self, bibtex_str, *args, **kwargs):

        order, indent = read_display_order(bibtex_str, kwargs.pop("tabsize", 2))
        data = bibtexparser.bparser.BibTexParser.parse(self, bibtex_str, *args, **kwargs)

        for key in order:
            order[key] = list(map(self.alt_dict.get, order[key], order[key]))

        for entry in data.entries:
            entry["INDENT"] = " " * indent
            entry["DISPLAY_ORDER"] = order.get(entry["ID"], [])

        return data


def parse(bibtex_str: str, aggresive: bool = False) -> str:
    """
    Parse a BibTeX string once.

    :param aggresive: Use aggressive interpretation strategy.
    """

    writer = MyBibTexWriter()

    if aggresive:
        parser = MyBibTexParser(
            homogenize_fields=True,
            ignore_nonstandard_types=True,
            add_missing_from_crossref=True,
            common_strings=True,
        )
    else:
        parser = MyBibTexParser()

    return writer.write(parser.parse(bibtex_str))


def _subr(pattern, repl, string):
    """
    Recursive replacement: apply ``re.subn`` until no more replacement is made.
    """

    string, nsub = re.subn(pattern, repl, string)

    if nsub:
        return _subr(pattern, repl, string)

    return string


def selection(use_bibtexparser: bool = False):
    """
    List of fields to keep in a BibTeX file to get a useful list of references:
    fields that are not in this selection may be useful for a database, but might only cloud
    BibTeX output.

    :param use_bibtexparser: Add bibtexparser specific fields to select (not part of BibTeX output).
    """

    base = []

    if use_bibtexparser:
        base += ["ID", "ENTRYTYPE", "DISPLAY_ORDER", "INDENT"]

    base += ["author", "title", "year", "doi", "arxivid"]

    return dict(
        article=base + ["journal", "volume", "number", "pages"],
        unpublished=base,
        inproceedings=base + ["booktitle", "editor", "publisher", "volume", "number", "pages"],
        book=base + ["edition", "editor", "publisher", "isbn", "volume", "pages"],
        inbook=base + ["edition", "editor", "publisher", "isbn", "volume", "pages"],
        incollection=base + ["booktitle", "edition", "publisher", "isbn", "volume", "pages"],
        phdthesis=base + ["school", "isbn", "url"],
        techreport=base + ["institution", "isbn", "url"],
        misc=base + ["pages", "url"],
    )


@singledispatch
def select(
    data,
    fields: Union[dict[list[str]], list[str]] = None,
    ensure_link: bool = True,
    remove_url: bool = True,
):
    """
    Remove unnecessary fields for BibTex file.

    :param data:
        The BibTeX database (file, string, or bibtexparser instance).

    :param fields:
        Fields to keep per entry type (default from :py:fund:`selection`).
        If a list is specified all entry types are treated the same.

    :param ensure_link:
        Add URL to ``fields`` if no ``doi`` or ``arxivid`` is present.

    :param remove_url:
        Remove URL with either a ``doi`` or an ``arxivid`` is present.
    """

    if fields is None:
        fields = selection(use_bibtexparser=True)

    if isinstance(fields, list):
        ret = {}
        for entry in data.entries:
            if entry["ENTRYTYPE"] not in ret:
                ret[entry["ENTRYTYPE"]] = ["ID", "ENTRYTYPE", "DISPLAY_ORDER", "INDENT"] + fields
        fields = ret

    for entry in data.entries:

        select = fields[entry["ENTRYTYPE"]]

        if ensure_link:
            if "url" not in select:
                if "doi" not in entry and "arxivid" not in entry:
                    select.append("url")

        rm = [key for key in entry if key not in select]
        for key in rm:
            del entry[key]

        if remove_url:
            if "url" in entry and ("doi" in entry or "arxivid" in entry):
                del entry["url"]

    return data


@select.register(str)
def _(data, *args, **kwargs):

    writer = MyBibTexWriter()
    parser = MyBibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=True,
        add_missing_from_crossref=True,
        common_strings=True,
    )
    return writer.write(select(parser.parse(data), *args, **kwargs))


@select.register(io.IOBase)
def _(data, *args, **kwargs):

    writer = MyBibTexWriter()
    parser = MyBibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=True,
        add_missing_from_crossref=True,
        common_strings=True,
    )

    return writer.write(select(parser.parse(data), *args, **kwargs))


def unique(data: bibtexparser.bibdatabase.BibDatabase, merge: bool = True):
    """
    Remove duplicate keys.
    :param data: The BibTeX database.
    :param merge: Try to keep as many keys as possible.
    :return: The BibTeX database.
    """

    keys = [entry["ID"] for entry in data.entries]
    _, ifoward, ibackward = np.unique(keys, return_index=True, return_inverse=True)

    if ifoward.size == len(keys):
        return data

    index = np.arange(ibackward.size)
    renum = index[ifoward][ibackward]
    old = index[index != renum]
    new = renum[index != renum]

    entries = [data.entries[i] for i in ifoward]
    merged = []

    for o, n in zip(old, new):
        merged.append(data.entries[o]["ID"])
        if merge:
            for key in data.entries[o]:
                if key not in data.entries[n]:
                    entries[n][key] = data.entries[o][key]

    sorter = np.argsort(ifoward)
    data.entries = [entries[i] for i in sorter]

    merged = "- " + "\n- ".join([str(i) for i in np.unique(merged)])
    warnings.warn(f"Merging duplicates, please check:\n{merged}", Warning)

    return data


@singledispatch
def clean(
    data: bibtexparser.bibdatabase.BibDatabase,
    journal_type: str = "abbreviation",
    journal_database: list[str] = ["pnas", "physics", "mechanics"],
    sep_name: str = "",
    sep_journal: str = "",
    title: bool = True,
    protect_math: bool = True,
    rm_unicode: bool = True,
    rm_comment: bool = True,
    rm_string: bool = True,
) -> bibtexparser.bibdatabase.BibDatabase:
    r"""
    Clean a BibTeX database:
    *   Remove unnecessary fields (see :py:func:`GooseBib.bibtex.select`).
    *   Unify the formatting of authors (see :py:func:`GooseBib.reformat.abbreviate_firstname`).
    *   Ensure proper math formatting (see :py:func:`GooseBib.reformat.protect_math`).
    *   Convert unicode to TeX (see :py:func:`GooseBib.reformat.rm_unicode`).
    *   Fill digital identifier if it is not present by can be found from a different field
        (see :py:func:`GooseBib.recognise.doi` and :py:func:`GooseBib.recognise.arxivid`).

    :param data: The BibTeX database (file, string, or bibtexparser instance).
    :param journal_type: Use journal: "title", "abbreviation", or "acronym".
    :param journal_database: Database(s) with official journal names, abbreviations, and acronym.
    :param sep_name: Separate name initials (e.g. "", " ").
    :param sep_journal: Separate abbreviations in journal: replace ". " e.g. by ". " or ".\ ".
    :param title: Include title of relevant fields.
    :param protect_math: Apply fix of :py:func:`reformat.protect_math`.
    :param rm_unicode: Apply fix of :py:func:`reformat.rm_unicode`.
    :param rm_comment: Remove @comment.
    :param rm_string: Remove @string (as it is interpreted by the parser).
    :param sort_entries: Sort entries in output (only for string or file input).
    """

    data = unique(data, merge=True)

    if rm_comment:
        data.comments = []

    if rm_string:
        data.strings = OrderedDict()

    journal_type = journal_type.lower()
    journal_database = [journal_database] if isinstance(journal_database, str) else journal_database

    revus = []
    ignored_authors = []

    for entry in data.entries:

        # prepare journal rename
        if "journal" in entry:
            revus.append(entry["journal"])

        # find doi
        if "doi" not in entry:
            doi = recognise.doi(
                *[
                    val
                    for key, val in entry.items()
                    if key not in ["arxivid", "eprint", "DISPLAY_ORDER", "INDENT"]
                ]
            )
            if doi:
                entry["doi"] = doi

        # find arXiv-id
        if "arxivid" not in entry:
            arxivid = recognise.arxivid(
                *[
                    val
                    for key, val in entry.items()
                    if key not in ["doi", "DISPLAY_ORDER", "INDENT"]
                ]
            )
            if arxivid:
                entry["arxivid"] = arxivid

        # apply arXiv's doi
        if "arxivid" in entry:
            if "doi" not in entry:
                entry["doi"] = "10.48550/arXiv." + entry.pop("arxivid")
            elif entry["doi"] == "10.48550/arXiv." + entry["arxivid"]:
                del entry["arxivid"]

        # fix author abbreviations
        for key in ["author", "editor"]:
            if key in entry:
                names = re.split(r"\ and\ ", entry[key].replace("\n", " "), flags=re.IGNORECASE)
                if not re.match(r"(\{)(.*)(\})", entry[key]):
                    names = bibtexparser.customization.getnames(names)
                names = [reformat.abbreviate_firstname(i, sep_name) for i in names]
                entry[key] = " and ".join(names)

        # remove title
        if not title:
            entry.pop("title", None)

        # protect math-mode
        if protect_math:
            for key in ["title"]:
                if key in entry:
                    entry[key] = reformat.protect_math(entry[key])

        # convert unicode to LaTeX
        if rm_unicode:
            for key in ["author", "editor", "title"]:
                if key in entry:
                    entry[key] = reformat.rm_unicode(entry[key])

        # abbreviations: change symbol after "."
        for key in ["journal"]:
            if key in entry:
                entry[key] = entry[key].replace(". ", f".{sep_journal} ")
                entry[key] = entry[key].replace(r".\ ", f".{sep_journal} ")

        # fix underscore problems
        # -
        if "doi" in entry:
            entry["doi"] = entry["doi"].replace("_", r"\_")
            entry["doi"] = re.sub(r"[\{}]?[\\]+\_[\}]?", r"\\_", entry["doi"])
        # -
        if "url" in entry:
            entry["url"] = entry["url"].replace(r"{\_}", r"\_")
            entry["url"] = re.sub(r"[\{}]?[\\]+\_[\}]?", r"\\_", entry["url"])
            entry["url"] = entry["url"].replace("{~}", "~")
            entry["url"] = entry["url"].replace(r"\&", "&")
        # -
        if "url" in entry:
            entry["url"] = _subr(re.compile(r"({)([^}])(})", re.UNICODE), r"\2", entry["url"])

    if len(ignored_authors) > 0:
        ignored_authors = "- " + "\n- ".join([str(i) for i in np.unique(ignored_authors)])
        warnings.warn(f"Protected authors found, please check:\n{ignored_authors}", Warning)

    # rename journal

    if len(journal_database) > 0:

        db = journals.load(*journal_database)
        if journal_type in ["title", "name", "official", "off"]:
            new = db.map2name(revus)
        elif journal_type in ["abbreviation", "abbr"]:
            new = db.map2abbreviation(revus)
        elif journal_type in ["acronym", "acro"]:
            new = db.map2acronym(revus)
        else:
            raise OSError(f'Unknown journal type selection "{journal_type}"')

        mapping = {o: n for o, n in zip(revus, new)}

        for entry in data.entries:
            if "journal" in entry:
                entry["journal"] = mapping[entry["journal"]]

    # return selection of fields

    return select(data, fields=selection(use_bibtexparser=True))


@clean.register(str)
def _(data: str, *args, **kwargs) -> str:

    writer = MyBibTexWriter(sort_entries=kwargs.pop("sort_entries", False))
    parser = MyBibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=True,
        add_missing_from_crossref=True,
        common_strings=True,
    )
    return writer.write(clean(parser.parse(data), *args, **kwargs))


@clean.register(io.IOBase)
def _(data: io.IOBase, *args, **kwargs) -> str:

    writer = MyBibTexWriter(sort_entries=kwargs.pop("sort_entries", False))
    parser = MyBibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=True,
        add_missing_from_crossref=True,
        common_strings=True,
    )
    return writer.write(clean(parser.parse(data), *args, **kwargs))


@singledispatch
def format_journal_arxiv(data, fmt: str, journal_database: list[str] = ["arxiv"]):
    """
    Format the journal entry for arXiv preprints.
    Use "{}" in the formatter to include the arxivid.

    :param data: The BibTeX database (file, string, or bibtexparser instance).
    :param fmt: Formatter, e.g. "Preprint" or "Preprint: arXiv {}".
    :param journal_database: Database(s) with known arXiv variants.
    """

    search = r"(" + re.escape("10.48550/arXiv.") + r")(.*)"
    pattern = ["arxiv", "preprint", "submitted", "in preparation"]

    for entry in data.entries:

        if "doi" in entry:
            if not re.match(search, entry["doi"]):
                continue

        if "arxivid" in entry:
            arxivid = entry["arxivid"]
        elif "doi" in entry:
            arxivid = re.split(search, entry["doi"])
            if len(arxivid) != 4:
                continue
            arxivid = arxivid[2]
        else:
            continue

        if "journal" not in entry:
            entry["journal"] = fmt.format(arxivid)
        else:
            for i in pattern:
                if i in entry["journal"].lower():
                    entry["journal"] = fmt.format(arxivid)
                    break

    if len(journal_database) > 0:

        revus = []
        for entry in data.entries:
            if "journal" in entry:
                revus.append(entry["journal"])

        db = journals.load(*journal_database)
        new = db.map2name(revus)
        mapping = {o: n for o, n in zip(revus, new)}

        for entry in data.entries:
            if "journal" in mapping and "arxivid" in entry:
                if "doi" in entry:
                    if not re.match(search, entry["doi"]):
                        continue
                entry["journal"] = fmt.format(arxivid)

    return data


@format_journal_arxiv.register(str)
def _(data, *args, **kwargs):

    writer = MyBibTexWriter()
    parser = MyBibTexParser()
    return writer.write(format_journal_arxiv(parser.parse(data), *args, **kwargs))


@format_journal_arxiv.register(io.IOBase)
def _(data, *args, **kwargs):

    writer = MyBibTexWriter()
    parser = MyBibTexParser()
    return writer.write(format_journal_arxiv(parser.parse(data), *args, **kwargs))


def GbibClean():
    r"""
    Clean a BibTeX database, stripping it from unnecessary fields,
    unifying the formatting of authors,
    and ensuring the proper special characters and math mode settings.
    This script preserves order to help tracking changes.

    :usage:

        GbibClean [options] <input>... <output>

    :arguments:

        <input>
            Input BibTeX-file(s).

        <output>
            Output file.
            If ``<output>`` is a directory, it is appended with the (first) filename of ``<input>``.
            Multiple input files are combined to a single output file, in every way they are
            considered as concatenated.
            See ``--in-place`` for different behaviour.

    :options:

        -o, --output=STR
            Specify output file. In this case all arguments are interpreted as input files
            (i.e. also the last argument).

        -j, --journal-type=STR (title, abbreviation, acronym)
            Unify journal titles (if recognised).
            If "acronym" is selected, "abbreviation" is used for journals without an "acronym".
            Default: abbreviation.

        --journals=STR (physics, mechanics, arxiv, pnas, pnas-usa, ...)
            Database(s) with official journal names, abbreviations, and acronyms.
            To make no modifications to journals use ``--journal=""``.
            Default: "pnas,physics,mechanics".

        --arxiv=STR
            Format arXiv preprints using a specific formatter.
            Use "{}" in the formatter to include the arxivid, e.g.: "arXiv preprint {}".

        --no-title
            Remove title from BibTeX file.

        --author-sep=STR
            Character to separate authors' initials.
            Default: "".

        --journal-sep=<str>
            Separate journal abbreviations by ".{sep} "
            Default: "".

        --ignore-case
            Do not protect case of title.

        --ignore-math
            Do not apply math-mode fix.

        --ignore-unicode
            Do not apply unicode fix.

        --sort-entries
            Sort output by entries.

        --diff=STR
            Write diff to HTML file which shows the old and the reformatted file side-by-side.
            See ``difflib.HtmlDiff.make_file``.

        --diff-context=BOOL
            Show contextual differences.
            See ``difflib.HtmlDiff.make_file``.

        --diff-numlines=BOOL
            Controls the number of context lines which surround the difference highlights.

        --diff-type=STR (raw, plain, all, select)
            Show difference between ``<output>`` and compared to ``<input>`` that is:
            *   raw: not parsed at all (can lead to a mess because of sorting).
            *   plain: parsed as little as possible.
            *   select: parsed as little as possible, but with only selected output fields.
            Default: select.

        --diff-keys=STR
            Limit diff to certain keys separated by spaces (e.g. "author,journal,doi").

        --in-place
            If specified, each input file is separately treated and formatted in-place.

        -f, --force
            Force overwrite of existing files.

        -v, --version
            Show version.

        -h, --help
            Show help.

    (c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseBib
    """

    funcname = inspect.getframeinfo(inspect.currentframe()).function
    doc = textwrap.dedent(inspect.getdoc(globals()[funcname]))

    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(doc)

    parser = Parser()
    parser.add_argument("--arxiv", type=str)
    parser.add_argument("--author-sep", type=str, default="")
    parser.add_argument("--sort-entries", action="store_true")
    parser.add_argument("--diff", type=str)
    parser.add_argument("--diff-type", type=str, default="select")
    parser.add_argument("--diff-keys", type=str)
    parser.add_argument("--diff-context", type=bool, default=False)
    parser.add_argument("--diff-numlines", type=int, default=5)
    parser.add_argument("--ignore-case", action="store_true")
    parser.add_argument("--ignore-math", action="store_true")
    parser.add_argument("--ignore-unicode", action="store_true")
    parser.add_argument("--in-place", action="store_true")
    parser.add_argument("--journal-sep", type=str, default="")
    parser.add_argument("--journals", type=str, default="pnas,physics,mechanics")
    parser.add_argument("--no-title", action="store_true")
    parser.add_argument("-o", "--output", type=str)
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("-j", "--journal-type", type=str, default="abbreviation")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("files", nargs="*", type=str)
    args = parser.parse_args()

    # read input/output filepaths

    if args.in_place:

        assert args.output is None
        assert not args.diff
        assert not args.force

        sourcepaths = args.files
        outpaths = sourcepaths
        assert all([os.path.isfile(i) for i in sourcepaths])

    else:

        if args.output is not None:
            if len(args.files) < 1:
                raise OSError("Specify <input>")
            output = args.output
            sources = args.files
        else:
            if len(args.files) < 2:
                raise OSError("Specify <input> and <output>")
            output = args.files[-1]
            sources = args.files[:-1]

        source = ""

        if os.path.isdir(output):
            output = os.path.join(output, os.path.split(sources[0])[-1])

        sourcepaths = [None]
        outpaths = [output]

        for filepath in sources:
            if not os.path.isfile(filepath):
                raise OSError(f'"{filepath}" does not exist')
            with open(filepath) as file:
                source += file.read()

        if not args.force:

            overwrite = []

            for outpath in outpaths:
                if os.path.isfile(outpath):
                    overwrite += [os.path.normpath(outpath)]

            if args.diff:
                if os.path.isfile(args.diff):
                    overwrite += [os.path.normpath(args.diff)]

            if len(overwrite) > 0:
                files = ", ".join(overwrite)
                if not click.confirm(f'Overwrite "{files}"?'):
                    raise OSError("Cancelled")

    # formatting

    for sourcepath, outpath in zip(sourcepaths, outpaths):

        if sourcepath is not None:
            with open(sourcepath) as file:
                source = file.read()

        data = clean(
            source,
            journal_type=args.journal_type,
            journal_database=args.journals.split(","),
            sep_name=args.author_sep,
            sep_journal=args.journal_sep,
            title=not args.no_title,
            protect_math=not args.ignore_math,
            rm_unicode=not args.ignore_unicode,
            sort_entries=args.sort_entries,
        )

        if data != parse(data):
            warnings.warn("Re-parsing is failing, there might be dangling {}", Warning)

        if args.arxiv:
            data = format_journal_arxiv(data, args.arxiv)

        if data == source:
            return 0

        with open(outpath, "w") as file:
            file.write(data)

        if args.diff is not None:

            if args.diff_type.lower() == "raw":
                simple = source
            elif args.diff_type.lower() == "plain":
                try:
                    simple = parse(source)
                except:
                    simple = parse(source, aggresive=True)
                    warnings.warn("Light parsing for diff failed", Warning)
            elif args.diff_type.lower() == "select":
                simple = select(source)
            else:
                raise OSError("Unknown option for --diff-type")

            if args.diff_keys:
                simple = select(
                    simple,
                    fields=args.diff_keys.split(","),
                    ensure_link=False,
                    remove_url=False,
                )
                data = select(
                    data,
                    fields=args.diff_keys.split(","),
                    ensure_link=False,
                    remove_url=False,
                )

            diff = difflib.HtmlDiff(wrapcolumn=100).make_file(
                simple.splitlines(keepends=True),
                data.splitlines(keepends=True),
                numlines=args.diff_numlines,
                context=args.diff_context,
            )

            with open(args.diff, "w") as file:
                file.write(diff)


def GbibShowAuthorRename():
    r"""
    Show author rename if ``GbibClean`` is applied.
    """

    funcname = inspect.getframeinfo(inspect.currentframe()).function
    doc = textwrap.dedent(inspect.getdoc(globals()[funcname]))

    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(doc)

    parser = Parser()
    parser.add_argument("-o", "--output", type=str)
    parser.add_argument("--author-sep", type=str, default="")
    parser.add_argument("--all", action="store_true")
    parser.add_argument("files", nargs="+", type=str)
    args = parser.parse_args()

    assert all([os.path.isfile(i) for i in args.files])

    source = ""

    for filepath in args.files:
        with open(filepath) as file:
            source += file.read()

    parser = MyBibTexParser(
        homogenize_fields=True,
        ignore_nonstandard_types=True,
        add_missing_from_crossref=True,
        common_strings=True,
    )

    data = parser.parse(source)
    old = []
    new = []

    for entry in data.entries:
        for key in ["author", "editor"]:
            if key in entry:
                names = re.split(r"\ and\ ", entry[key].replace("\n", " "), flags=re.IGNORECASE)
                old += names
                if not re.match(r"(\{)(.*)(\})", entry[key]):
                    names = bibtexparser.customization.getnames(names)
                new += [reformat.abbreviate_firstname(i, args.author_sep) for i in names]

    _, index = np.unique(old, return_index=True)
    old = [old[i] for i in index]
    new = [new[i] for i in index]

    if args.all:
        opts = dict(context=False, numlines=1)
    else:
        opts = dict(context=True, numlines=0)

    diff = difflib.HtmlDiff(wrapcolumn=100).make_file(old, new, **opts)

    if args.output:
        with open(args.output, "w") as file:
            file.write(diff)
    else:
        print(diff)
